
/*
 * The Real SoundTracker - gtk+ Playlist widget
 *
 * Copyright (C) 1999-2019 Michael Krause
 * Copyright (C) 2020-2022 Yury Aliaev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

#include "extspinbutton.h"
#include "gui-subs.h"
#include "history.h"
#include "marshal.h"
#include "playlist.h"
#include "st-subs.h"

/* --- Structures, variables, definitions --- */

struct _PlaylistPriv {
    GtkWidget *mainwindow, *dialog;
    GtkAdjustment* adj_songpos;
    GtkWidget *spin_songlength, *spin_songpat, *spin_restartpos, *spin_editor_pat;
    GtkWidget *ibutton, *icbutton, *ifbutton, *dbutton;
    GtkWidget *list, *label_editor_ord;
    GtkListStore* ls;
    GtkTreeSelection* sel;
    GtkTreeModel* tm;

    GtkWidget *numlabels[5], *patlabels[5];

    gint max_length;
    gint min_pattern;
    gint max_pattern;

    gint current_position;
    gint signals_disabled;
    gboolean frozen, enabled;
    gint rbutton_tag, tag_del, tag_ins;
    gint skip_logging;
    gint row_ins;

    XM* xm;
};

enum {
    SIG_CURRENT_POSITION_CHANGED,
    SIG_ENTRY_CHANGED,
    SIG_INSERT_PATTERN,
    SIG_ADD_COPY,
    LAST_SIGNAL
};

static guint playlist_signals[LAST_SIGNAL] = { 0 };

/* --- Utility functions --- */

static void
playlist_update_adjustment(PlaylistPriv* p,
    const gint min,
    const gint max)
{
    gint newpos = CLAMP(p->current_position, min, max - 1);

    g_object_freeze_notify(G_OBJECT(p->adj_songpos));
    gtk_adjustment_set_lower(p->adj_songpos, min);
    gtk_adjustment_set_upper(p->adj_songpos, max);
    if (newpos != p->current_position)
        gtk_adjustment_set_value(p->adj_songpos, newpos);
    g_object_thaw_notify(G_OBJECT(p->adj_songpos));
}

static void
playlist_draw_contents(PlaylistPriv* p,
    const gboolean update_playlist,
    const gboolean update_editor)
{
    gint i;
    gchar buf[4];

    g_assert(p->xm != NULL);

    if (p->dialog) {
        g_snprintf(buf, sizeof(buf), "%u", p->current_position);
        gtk_label_set_text(GTK_LABEL(p->label_editor_ord), buf);
    }

    if (update_playlist) {
        for (i = 0; i <= 4; i++) {
            if ((i >= 2 - p->current_position) &&
                (i <= p->xm->song_length - p->current_position + 1)) {
                g_snprintf(buf, sizeof(buf), "%.3u", i - 2 + p->current_position);
                gtk_label_set_text(GTK_LABEL(p->numlabels[i]), buf);
                if (i != 2) {
                    g_snprintf(buf, sizeof(buf), "%.3u", p->xm->pattern_order_table[i - 2 + p->current_position]);
                    gtk_label_set_text(GTK_LABEL(p->patlabels[i]), buf);
                }
            } else {
                gtk_label_set_text(GTK_LABEL(p->numlabels[i]), "");
                gtk_label_set_text(GTK_LABEL(p->patlabels[i]), "");
            }
        }
    }
    if (update_editor && p->dialog) {
        GtkTreeIter iter;

        g_signal_handler_block(G_OBJECT(p->tm), p->tag_del);
        g_signal_handler_block(G_OBJECT(p->tm), p->tag_ins);
        gui_list_freeze(p->list);
        gui_list_clear_with_model(p->tm);
        gtk_tree_view_set_model(GTK_TREE_VIEW(p->list), NULL);
        for (i = 0; i < p->xm->song_length; i++) {
            gtk_list_store_append(p->ls, &iter);
            gtk_list_store_set(p->ls, &iter, 0, i,
                1, p->xm->pattern_order_table[i],
                2, p->xm->patterns[p->xm->pattern_order_table[i]].length, -1);
        }
        gui_list_thaw(p->list, p->tm);
        g_signal_handler_unblock(G_OBJECT(p->tm), p->tag_del);
        g_signal_handler_unblock(G_OBJECT(p->tm), p->tag_ins);
    }
}

static inline void
playlist_set_length(PlaylistPriv* p, const gint length)
{
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(p->spin_songlength), length);
}

static inline void
playlist_set_restartpos(PlaylistPriv* p, const gint pos)
{
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(p->spin_restartpos), pos);
}

struct PlaylistUndoArg {
    gint length, pos, restartpos, src, dest;
    guint8 pot[256];
};

static void
playlist_undo(const gint smp, const gint ins, const gboolean redo,
    gpointer arg, gpointer data)
{
    gint tmp_value, old_restartpos;
    Playlist* p;
    PlaylistPriv* priv;
    struct PlaylistUndoArg* state = arg;
    guint8* pot = alloca(sizeof(priv->xm->pattern_order_table));

    g_assert(IS_PLAYLIST(data));
    p = PLAYLIST(data);
    priv = p->priv;
    g_assert(priv->xm);

    memcpy(pot, priv->xm->pattern_order_table, sizeof(priv->xm->pattern_order_table));
    tmp_value = priv->xm->song_length;
    old_restartpos = priv->xm->restart_position;
    playlist_freeze(p);
    playlist_freeze_signals(p);
    playlist_set_length(priv, state->length);
    state->length = tmp_value;
    playlist_set_restartpos(priv, state->restartpos);
    state->restartpos = old_restartpos;
    memcpy(priv->xm->pattern_order_table, state->pot, sizeof(priv->xm->pattern_order_table));
    memcpy(state->pot, pot, sizeof(state->pot));
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songpat),
        priv->xm->pattern_order_table[priv->current_position]);
    tmp_value = priv->current_position;
    gtk_adjustment_set_value(priv->adj_songpos, state->pos);
    state->pos = tmp_value;

    if (state->src >= 0) {
        if (redo)
            st_copy_pattern(&priv->xm->patterns[state->dest], &priv->xm->patterns[state->src]);
        else
            st_clear_pattern(&priv->xm->patterns[state->dest]);
    }

    if (!priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_CURRENT_POSITION_CHANGED],
            0, tmp_value);
    playlist_thaw_signals(p);
    playlist_thaw(p);
}

static void
playlist_log(Playlist* p,
    const gchar* const title,
    const gint src,
    const gint dest)
{
    PlaylistPriv* priv = p->priv;
    struct PlaylistUndoArg* arg = g_new(struct PlaylistUndoArg, 1);

    arg->length = priv->xm->song_length;
    arg->pos = priv->current_position;
    arg->restartpos = priv->xm->restart_position;
    arg->src = src,
    arg->dest = dest;
    memcpy(arg->pot, priv->xm->pattern_order_table, sizeof(arg->pot));
    history_log_action(HISTORY_ACTION_POINTER, _(title),
        HISTORY_FLAG_LOG_PAT, playlist_undo, p, sizeof(struct PlaylistUndoArg), arg);
}
static void
_playlist_insert_pattern(Playlist* p,
    const gchar* const title,
    const gint pos,
    const gint pat,
    const gint src)
{
    gint current_songpos;
    PlaylistPriv* priv = p->priv;
    gint length;

    g_assert(IS_PLAYLIST(p));
    g_assert(priv->xm != NULL);
    length = priv->xm->song_length;
    g_assert(pos >= 0 && pos <= length);
    g_assert(pat >= priv->min_pattern && pat <= priv->max_pattern);

    if (length == priv->max_length)
        return;

    current_songpos = lrint(gtk_adjustment_get_value(priv->adj_songpos));

    playlist_log(p, _(title), src, pat);

    playlist_freeze(p);
    playlist_freeze_signals(p);
    priv->skip_logging++;
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songlength), length + 1);
    priv->skip_logging--;
    playlist_thaw_signals(p);

    memmove(&priv->xm->pattern_order_table[pos + 1], &priv->xm->pattern_order_table[pos],
        sizeof(priv->xm->pattern_order_table) - (1 + pos) * sizeof(priv->xm->pattern_order_table[0]));
    priv->xm->pattern_order_table[pos] = pat;
    if (pos == current_songpos) {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songpat),
            priv->xm->pattern_order_table[pos]);
    }

    /* This also redraws the playlist */
    playlist_thaw(p);
}

static void editor_select_current_row(PlaylistPriv* p)
{
    GtkTreeIter sel_iter;

    if (p->dialog && gui_list_get_iter(p->current_position, p->ls, &sel_iter)) {
        gtk_tree_selection_select_iter(p->sel, &sel_iter);
        gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(p->list),
            gtk_tree_model_get_path(p->tm, &sel_iter), NULL, TRUE, 0.5, 0.5);
    }
}
/* --- GUI callbacks --- */

static void
editor_row_activated(GtkTreeView* tv,
    GtkTreePath* path,
    GtkTreeViewColumn* column,
    Playlist* p)
{
    gint* indices = gtk_tree_path_get_indices(path);

    if (indices[0] != p->priv->current_position)
        playlist_set_position(p, indices[0]);
}

static void
pot_undo(const gint smp, const gint ins, const gboolean redo,
    gpointer arg, gpointer data)
{
    PlaylistPriv* p = data;
    guint8 *pot = arg, *tmp = alloca(sizeof(p->xm->pattern_order_table));

    memcpy(tmp, p->xm->pattern_order_table, sizeof(p->xm->pattern_order_table));
    memcpy(p->xm->pattern_order_table, pot, sizeof(p->xm->pattern_order_table));
    memcpy(pot, tmp, sizeof(p->xm->pattern_order_table));
    playlist_draw_contents(p, TRUE, TRUE);
    editor_select_current_row(p);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(p->spin_songpat),
        p->xm->pattern_order_table[p->current_position]);
}

static void
playlist_log_pot(PlaylistPriv* p)
{
    guint8* pot = g_malloc(sizeof(p->xm->pattern_order_table));

    memcpy(pot, p->xm->pattern_order_table, sizeof(p->xm->pattern_order_table));
    history_log_action(HISTORY_ACTION_POINTER, _("Pattern sequence modification"),
        HISTORY_FLAG_LOG_POS | HISTORY_FLAG_LOG_PAT,
        pot_undo, p, sizeof(p->xm->pattern_order_table), pot);
}

static void
editor_row_deleted(GtkTreeView* tv, GtkTreePath* path, Playlist* pl)
{
    PlaylistPriv* p = pl->priv;
    gint* indices = gtk_tree_path_get_indices(path);
    gint row_del = indices[0], row_ins = p->row_ins;
    guint8 tmp;

    g_assert(p->xm != NULL);
    /* Something's wrong, just ignoring */
    if (p->row_ins == -1)
        return;

    /* Somewhat strange logic of treeview dnd */
    if (row_ins < row_del - 1) {
        playlist_log_pot(p);
        tmp = p->xm->pattern_order_table[row_del - 1];
        memmove(&p->xm->pattern_order_table[row_ins + 1],
                &p->xm->pattern_order_table[row_ins],
                row_del - row_ins - 1);
        p->xm->pattern_order_table[row_ins] = tmp;
        if (p->current_position != row_ins) {
            playlist_set_position(pl, row_ins);
            playlist_draw_contents(p, FALSE, TRUE);
        } else {
            playlist_draw_contents(p, TRUE, TRUE);
        }
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(p->spin_songpat), tmp);
        editor_select_current_row(p);
    } else if (row_ins > row_del + 1) {
        playlist_log_pot(p);
        tmp = p->xm->pattern_order_table[row_del];
        memmove(&p->xm->pattern_order_table[row_del],
                &p->xm->pattern_order_table[row_del + 1],
                row_ins - row_del - 1);
        p->xm->pattern_order_table[row_ins - 1] = tmp;
        if (p->current_position != row_ins - 1) {
            playlist_set_position(pl, row_ins - 1);
            playlist_draw_contents(p, FALSE, TRUE);
        } else {
            playlist_draw_contents(p, TRUE, TRUE);
        }
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(p->spin_songpat), tmp);
        editor_select_current_row(p);
    }

    p->row_ins = -1;
}

static void
editor_row_inserted(GtkTreeView* tv, GtkTreePath* path, GtkTreeIter* itr, PlaylistPriv* p)
{
    gint* indices = gtk_tree_path_get_indices(path);
    p->row_ins = indices[0];
}

static void
playlist_insert_clicked(Playlist* p)
{
    if (!p->priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_INSERT_PATTERN],
            0, p->priv->current_position);
}

static void
playlist_delete_clicked(Playlist* p)
{
    gint pos;
    PlaylistPriv* priv = p->priv;

    g_assert(priv->xm != NULL);

    if (priv->xm->song_length == 1)
        return;

    pos = lrint(gtk_adjustment_get_value(priv->adj_songpos));
    playlist_log(p, N_("Pattern removal"), -1, -1);
    memmove(&priv->xm->pattern_order_table[pos], &priv->xm->pattern_order_table[pos + 1],
        sizeof(priv->xm->pattern_order_table) - (1 + pos) * sizeof(priv->xm->pattern_order_table[0]));

    /* Length change is already logged */
    priv->skip_logging++;
    playlist_set_length(priv, priv->xm->song_length - 1);
    priv->skip_logging--;

    pos = lrint(gtk_adjustment_get_value(priv->adj_songpos));
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songpat),
        priv->xm->pattern_order_table[pos]);
    editor_select_current_row(priv);
    if (!priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_CURRENT_POSITION_CHANGED],
            0, pos);
}

static void
playlist_add_free_clicked(Playlist* p)
{
    playlist_add_pattern(p, -1);
}

static void
playlist_add_copy_clicked(Playlist* p)
{
    if (!p->priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_ADD_COPY], 0);
}

static gboolean
playlist_button_release(GtkWidget* wid, GdkEventButton* event, Playlist* pl)
{
    PlaylistPriv* p = pl->priv;

    const gchar* titles[] = {N_("N"), N_("Pat"), N_("Len")};
    GType types[] = {G_TYPE_INT, G_TYPE_INT, G_TYPE_INT};
    const gfloat alignments[] = {0.5, 0.5, 0.5};
    const gboolean expands[] = {FALSE, FALSE, FALSE};

    g_assert(p->xm != NULL);

    if (event->button != 3)
        return FALSE;

    if (!p->dialog) {
        GtkWidget *vbox, *hbox, *thing;
        gint w, h;

        p->dialog = gtk_dialog_new_with_buttons(_("Pattern Sequence Editor"), GTK_WINDOW(p->mainwindow), 0,
            GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
        gui_dialog_adjust(p->dialog, GTK_RESPONSE_CLOSE);
        gui_dialog_connect(p->dialog, NULL);
        gui_translate_keyevents(p->dialog, p->mainwindow, FALSE);

        vbox = gtk_dialog_get_content_area(GTK_DIALOG(p->dialog));
        gtk_box_set_spacing(GTK_BOX(vbox), 2);

        hbox = gtk_hbox_new(FALSE, 4);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
        thing = gtk_label_new(_("Order"));
        gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

        p->label_editor_ord = gtk_label_new("0");
        gui_get_pixel_size(p->label_editor_ord, "000", &w, &h);
        gtk_widget_set_size_request(p->label_editor_ord, w, -1);
        gtk_misc_set_alignment(GTK_MISC(p->label_editor_ord), 1.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), p->label_editor_ord, FALSE, FALSE, 0);
        /* Spacer */
        thing = gtk_label_new("");
        gtk_widget_set_size_request(thing, 10, -1);
        gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

        thing = gtk_label_new(_("Pattern"));
        gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
        p->spin_editor_pat = extspinbutton_new(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(p->spin_songpat)),
            1.0, 0, FALSE);
        gtk_box_pack_start(GTK_BOX(hbox), p->spin_editor_pat, FALSE, FALSE, 0);

        p->list = gui_list_in_scrolled_window_full(3, titles, vbox,
            types, alignments, expands, GTK_SELECTION_BROWSE, TRUE, TRUE,
            NULL, GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
        gtk_tree_view_set_reorderable(GTK_TREE_VIEW(p->list), TRUE);
        p->ls = GUI_GET_LIST_STORE(p->list);
        p->sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(p->list));
        p->tm = gtk_tree_view_get_model(GTK_TREE_VIEW(p->list));
        g_signal_connect(p->list, "row_activated",
            G_CALLBACK(editor_row_activated), pl);
        p->tag_del = g_signal_connect(gtk_tree_view_get_model(GTK_TREE_VIEW(p->list)),
            "row_deleted", G_CALLBACK(editor_row_deleted), pl);
        p->tag_ins = g_signal_connect_after(gtk_tree_view_get_model(GTK_TREE_VIEW(p->list)),
            "row_inserted", G_CALLBACK(editor_row_inserted), p);

        hbox = gtk_hbox_new(FALSE, 4);
        gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

        thing = gui_button(GTK_STOCK_ADD, G_CALLBACK(playlist_insert_clicked), pl, hbox, TRUE, FALSE);
        gtk_widget_set_tooltip_text(thing, _("Insert pattern"));
        thing = gui_button(GTK_STOCK_REMOVE, G_CALLBACK(playlist_delete_clicked), pl, hbox, TRUE, FALSE);
        gtk_widget_set_tooltip_text(thing, _("Delete pattern"));
        thing = gui_button(GTK_STOCK_NEW, G_CALLBACK(playlist_add_free_clicked), pl, hbox, TRUE, FALSE);
        gtk_widget_set_tooltip_text(thing, _("Add free pattern"));
        thing = gui_button(GTK_STOCK_COPY, G_CALLBACK(playlist_add_copy_clicked), pl, hbox, TRUE, FALSE);
        gtk_widget_set_tooltip_text(thing, _("Add free pattern and copy"));

        thing = gtk_hseparator_new();
        gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);

        gtk_widget_set_size_request(p->list, -1, 100);
        gtk_widget_show_all(p->dialog);
    }

    /* It's easier just to repopulate patterns list on each editor
       showing up rather than taking care of its actual state */
    playlist_draw_contents(p, FALSE, TRUE);
    editor_select_current_row(p);

    gtk_window_present(GTK_WINDOW(p->dialog));
    return TRUE;
}

static gboolean
scroll_event(GtkWidget* w, GdkEventScroll* event, PlaylistPriv* p)
{
    int pos = p->current_position;

    switch (event->direction) {
    case GDK_SCROLL_UP:
        pos--;
        if ((pos >= 0) && (pos) < p->xm->song_length)
            gtk_adjustment_set_value(p->adj_songpos, pos);
        return TRUE;
    case GDK_SCROLL_DOWN:
        pos++;
        if ((pos >= 0) && (pos) < p->xm->song_length)
            gtk_adjustment_set_value(p->adj_songpos, pos);
        return TRUE;
    default:
        break;
    }
    return FALSE;
}

static gboolean
label_clicked(GtkWidget* w, GdkEventButton* event, PlaylistPriv* p)
{
    int i;

    int pos = p->current_position;

    if (event->button == 1) {
        // find what label is clicked
        for (i = -2; i <= 2; i++) {
            if ((p->numlabels[i + 2] == GTK_BIN(w)->child) || (p->patlabels[i + 2] == GTK_BIN(w)->child)) {
                pos += i;
                if ((pos >= 0) && (pos) < p->xm->song_length)
                    gtk_adjustment_set_value(p->adj_songpos, pos);
                return (TRUE);
            }
        }
    }
    return (FALSE);
}

static void
playlist_songpos_changed(GtkAdjustment* adj,
    Playlist* p)
{
    gint newpos;
    PlaylistPriv* priv = p->priv;

    g_assert(priv->xm != NULL);

    newpos = lrint(gtk_adjustment_get_value(adj));
    if (newpos == priv->current_position)
        return;

    priv->current_position = newpos;

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songpat),
        priv->xm->pattern_order_table[newpos]);

    if (!priv->frozen)
        playlist_draw_contents(priv, TRUE, FALSE);

    if (priv->dialog && gtk_widget_get_visible(priv->dialog))
        editor_select_current_row(priv);

    if (!priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_CURRENT_POSITION_CHANGED],
            0, newpos);
}

struct LNR {
    gint length, restartpos;
};

static void
song_length_undo(const gint smp, const gint ins, const gboolean redo,
    gpointer arg, gpointer data)
{
    gint tmp_value;
    struct LNR* values = arg;
    PlaylistPriv* priv = data;

    tmp_value = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(priv->spin_songlength));
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songlength), values->length);
    values->length = tmp_value;
    tmp_value = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(priv->spin_restartpos));
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_restartpos), values->restartpos);
    values->restartpos = tmp_value;
}

static void
playlist_songlength_changed(GtkSpinButton* spin,
    PlaylistPriv* priv)
{
    gint newlen;

    g_assert(priv->xm != NULL);

    newlen = gtk_spin_button_get_value_as_int(spin);

    if (newlen == priv->xm->song_length)
        return;

    if (!priv->skip_logging && !history_test_collate(HISTORY_ACTION_POINTER, 0, priv)) {
            struct LNR* arg = g_new(struct LNR, 1);

            arg->length = priv->xm->song_length;
            arg->restartpos = priv->xm->restart_position;
            history_log_action(HISTORY_ACTION_POINTER, _("Song length changing"),
                0, song_length_undo, priv, sizeof(struct LNR), arg);
    }

    priv->xm->song_length = newlen;
    playlist_update_adjustment(priv, 0, newlen);
    priv->skip_logging++;
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(priv->spin_restartpos), 0, newlen - 1);
    priv->skip_logging--;

    if (!priv->frozen)
        playlist_draw_contents(priv, TRUE, TRUE);
}

static void
playlist_songpat_changed(GtkSpinButton* spin,
    Playlist* p)
{
    PlaylistPriv* priv = p->priv;
    gint n = gtk_spin_button_get_value_as_int(spin);

    g_assert(priv->xm != NULL);

    if (priv->xm->pattern_order_table[priv->current_position] == n)
        return;

    history_log_spin_button(spin, _("Pattern number changing"),
        HISTORY_FLAG_LOG_POS, priv->xm->pattern_order_table[priv->current_position]);
    priv->xm->pattern_order_table[priv->current_position] = n;
    if (!priv->signals_disabled)
        g_signal_emit(G_OBJECT(p),
            playlist_signals[SIG_ENTRY_CHANGED],
            0, n);
}

static void
playlist_restartpos_changed(GtkSpinButton* spin,
    PlaylistPriv* p)
{
    gint n = gtk_spin_button_get_value_as_int(spin);

    g_assert(p->xm != NULL);

    if (p->xm->restart_position == n)
        return;

    if (!p->skip_logging)
        history_log_spin_button(spin, _("Restart position changing"),
            0, p->xm->restart_position);
    p->xm->restart_position = n;
}

static void
is_realized(GtkWidget* algn, gpointer data)
{
    guint x, y;

    gtk_widget_realize(algn);
    x = (algn->allocation).width;
    y = (algn->allocation).height;
    gtk_widget_set_size_request(algn, x, y);
}

/* --- API functions --- */

GtkWidget*
playlist_new(GtkWidget* mainwindow)
{
    GtkWidget *box, *thing, *thing1, *vbox, *frame, *box1, *evbox, *al;
    GtkAdjustment* adj;
    gint i;

    Playlist *p = g_object_new(playlist_get_type(), NULL);
    PlaylistPriv *priv = p->priv;

    priv->mainwindow = mainwindow;
    priv->rbutton_tag = g_signal_connect(p, "button_release_event",
        G_CALLBACK(playlist_button_release), p);

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_container_add(GTK_CONTAINER(p), vbox);

    box = gtk_hbox_new(FALSE, 2);
    thing = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(thing), GTK_SHADOW_IN);
    thing1 = gtk_table_new(2, 5, FALSE);

    /* pattern list view */

    for (i = 0; i <= 4; i++)
        if (i != 2) {
            priv->numlabels[i] = gtk_label_new("");
            evbox = gtk_event_box_new();
            gtk_container_add(GTK_CONTAINER(evbox), priv->numlabels[i]);
            g_signal_connect(evbox, "button_press_event",
                G_CALLBACK(label_clicked), priv);
            box1 = gtk_hbox_new(FALSE, 0),
            gtk_box_pack_start(GTK_BOX(box1), evbox, FALSE, FALSE, 3);
            gtk_table_attach_defaults(GTK_TABLE(thing1), box1, 0, 1, i, i + 1);

            priv->patlabels[i] = gtk_label_new("");
            evbox = gtk_event_box_new();
            gtk_container_add(GTK_CONTAINER(evbox), priv->patlabels[i]);
            g_signal_connect(evbox, "button_press_event",
                G_CALLBACK(label_clicked), priv);
            box1 = gtk_hbox_new(FALSE, 0),
            gtk_box_pack_start(GTK_BOX(box1), evbox, FALSE, FALSE, 0);
            gtk_table_attach_defaults(GTK_TABLE(thing1), box1, 1, 2, i, i + 1);
        }
    /* central label */
    box1 = gtk_hbox_new(FALSE, 0);
    priv->numlabels[2] = gtk_label_new("");
    /* a bit trick to keep label size constant */
    al = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(al), priv->numlabels[2]);
    g_signal_connect(al, "realize", G_CALLBACK(is_realized), NULL);
    gtk_box_pack_start(GTK_BOX(box1), al, TRUE, TRUE, 0);

    /* current pattern */
    adj = GTK_ADJUSTMENT(gtk_adjustment_new(0, priv->min_pattern, priv->max_pattern, 1.0, 5.0, 0.0));
    priv->spin_songpat = extspinbutton_new(adj, 1.0, 0, TRUE);
    extspinbutton_disable_size_hack(EXTSPINBUTTON(priv->spin_songpat));
    g_signal_connect(priv->spin_songpat, "value-changed",
        G_CALLBACK(playlist_songpat_changed), p);
    gtk_box_pack_end(GTK_BOX(box1), priv->spin_songpat, TRUE, TRUE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
    gtk_container_add(GTK_CONTAINER(frame), box1);

    gtk_table_attach_defaults(GTK_TABLE(thing1), frame, 0, 2, 2, 3);

    evbox = gtk_event_box_new();
    gtk_container_add(GTK_CONTAINER(evbox), thing1);
    g_signal_connect(evbox, "button_press_event",
        G_CALLBACK(label_clicked), priv);
    g_signal_connect(evbox, "scroll-event",
        G_CALLBACK(scroll_event), priv);

    gtk_container_add(GTK_CONTAINER(thing), evbox);
    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);

    /* scrollbar */
    priv->adj_songpos = GTK_ADJUSTMENT(gtk_adjustment_new(priv->current_position,
        0.0, 1.0, 1.0, 5.0, 1.0));
    thing = gtk_vscrollbar_new(priv->adj_songpos);
    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);
    g_signal_connect(priv->adj_songpos, "value_changed",
        G_CALLBACK(playlist_songpos_changed), p);

    /* buttons */
    thing = gtk_vbox_new(TRUE, 0);

    thing1 = priv->ibutton = gtk_button_new_with_label(_("Insert"));
    gtk_widget_set_tooltip_text(thing1, _("Insert pattern that is being edited"));
    gtk_box_pack_start(GTK_BOX(thing), thing1, TRUE, FALSE, 0);
    g_signal_connect_swapped(thing1, "clicked",
        G_CALLBACK(playlist_insert_clicked), p);
    g_signal_connect(thing1, "button_release_event",
        G_CALLBACK(playlist_button_release), p);

    thing1 = priv->dbutton = gtk_button_new_with_label(_("Delete"));
    gtk_widget_set_tooltip_text(thing1, _("Remove current playlist entry"));
    gtk_box_pack_start(GTK_BOX(thing), thing1, TRUE, FALSE, 0);
    g_signal_connect_swapped(thing1, "clicked",
        G_CALLBACK(playlist_delete_clicked), p);
    g_signal_connect(thing1, "button_release_event",
        G_CALLBACK(playlist_button_release), p);

    thing1 = priv->icbutton = gtk_button_new_with_label(_("Add + Cpy"));
    gtk_widget_set_tooltip_text(thing1, _("Add a free pattern behind current position, and copy current pattern to it"));
    gtk_box_pack_start(GTK_BOX(thing), thing1, TRUE, FALSE, 0);
    g_signal_connect_swapped(thing1, "clicked",
        G_CALLBACK(playlist_add_copy_clicked), p);
    g_signal_connect(thing1, "button_release_event",
        G_CALLBACK(playlist_button_release), p);

    thing1 = priv->ifbutton = gtk_button_new_with_label(_("Add Free"));
    gtk_widget_set_tooltip_text(thing1, _("Add a free pattern behind current position"));
    gtk_box_pack_start(GTK_BOX(thing), thing1, TRUE, FALSE, 0);
    g_signal_connect_swapped(thing1, "clicked",
        G_CALLBACK(playlist_add_free_clicked), p);
    g_signal_connect(thing1, "button_release_event",
        G_CALLBACK(playlist_button_release), p);

    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 0);

    /* box with songlen and repstart spinbuttons */
    box = gtk_hbox_new(FALSE, 0);

    thing1 = gtk_label_new(_("Len"));
    gtk_box_pack_start(GTK_BOX(box), thing1, FALSE, FALSE, 0);

    adj = GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 1.0, priv->max_length, 1.0, 5.0, 0.0));
    thing1 = priv->spin_songlength = extspinbutton_new(adj, 1.0, 0, TRUE);
    extspinbutton_disable_size_hack(EXTSPINBUTTON(thing1));
    gtk_widget_set_tooltip_text(thing1, _("Song length"));
    gtk_box_pack_start(GTK_BOX(box), thing1, FALSE, FALSE, 2);
    g_signal_connect(thing1, "value-changed",
        G_CALLBACK(playlist_songlength_changed), priv);

    adj = GTK_ADJUSTMENT(gtk_adjustment_new(priv->current_position, 0.0, 1.0, 1.0, 5.0, 0.0));
    thing1 = priv->spin_restartpos = extspinbutton_new(adj, 1.0, 0, TRUE);
    extspinbutton_disable_size_hack(EXTSPINBUTTON(thing1));
    gtk_widget_set_tooltip_text(thing1, _("Song restart position"));
    gtk_box_pack_end(GTK_BOX(box), thing1, FALSE, FALSE, 0);
    g_signal_connect(thing1, "value-changed",
        G_CALLBACK(playlist_restartpos_changed), priv);

    thing1 = gtk_label_new(_("Rstrt"));
    gtk_box_pack_end(GTK_BOX(box), thing1, FALSE, FALSE, 2);

    gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 0);

    gtk_widget_show_all(GTK_WIDGET(p));
    return GTK_WIDGET(p);
}

void playlist_freeze(Playlist* p)
{
    g_assert(IS_PLAYLIST(p));
    p->priv->frozen++;
}

void playlist_thaw(Playlist* p)
{
    g_assert(IS_PLAYLIST(p));

    if (p->priv->frozen)
        p->priv->frozen--;
    if (!p->priv->frozen)
        playlist_draw_contents(p->priv, TRUE, TRUE);
}

void playlist_enable(Playlist* p,
    const gboolean enable)
{
    PlaylistPriv* priv;

    g_assert(IS_PLAYLIST(p));
    priv = p->priv;

    if (enable == priv->enabled)
        return;

    priv->enabled = enable;
    gtk_widget_set_sensitive(priv->spin_songlength, enable);
    gtk_widget_set_sensitive(priv->spin_songpat, enable);
    gtk_widget_set_sensitive(priv->spin_restartpos, enable);
    gtk_widget_set_sensitive(priv->ibutton, enable);
    gtk_widget_set_sensitive(priv->ifbutton, enable);
    gtk_widget_set_sensitive(priv->icbutton, enable);
    gtk_widget_set_sensitive(priv->dbutton, enable);
    enable ? g_signal_handler_unblock(G_OBJECT(p), priv->rbutton_tag) :
        g_signal_handler_block(G_OBJECT(p), priv->rbutton_tag);
    if (!enable && priv->dialog)
        gtk_widget_hide(priv->dialog);
}

void playlist_freeze_signals(Playlist* p)
{
    g_assert(IS_PLAYLIST(p));
    p->priv->signals_disabled++;
}

void playlist_thaw_signals(Playlist* p)
{
    g_assert(IS_PLAYLIST(p));
    g_assert(p->priv->signals_disabled > 0);

    p->priv->signals_disabled--;
}

void playlist_set_xm(Playlist* p, XM* xm)
{
    PlaylistPriv* priv;

    g_assert(IS_PLAYLIST(p));
    g_assert(xm != NULL);
    priv = p->priv;

    priv->xm = xm;

    playlist_freeze(p);
    playlist_freeze_signals(p);
    gtk_adjustment_set_value(priv->adj_songpos, 0);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songlength),
        priv->xm->song_length);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_restartpos),
        priv->xm->restart_position);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(priv->spin_songpat),
        priv->xm->pattern_order_table[0]);
    playlist_update_adjustment(priv, 0, xm->song_length);
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(priv->spin_restartpos), 0, xm->song_length - 1);
    playlist_thaw_signals(p);
    playlist_thaw(p);
}

gint playlist_get_nth_pattern(Playlist* p,
    const int pos)
{
    g_assert(IS_PLAYLIST(p));
    g_assert(p->priv->xm != NULL);
    g_assert(pos >= 0 && pos < p->priv->xm->song_length);

    return p->priv->xm->pattern_order_table[pos];
}

void
playlist_set_position(Playlist* p,
    const gint pos)
{
    g_assert(IS_PLAYLIST(p));
    g_assert(p->priv->xm != NULL);
    g_assert(pos >= 0 && pos < p->priv->xm->song_length);

    gtk_adjustment_set_value(p->priv->adj_songpos, pos);
}

gint playlist_get_position(Playlist* p)
{
    g_assert(IS_PLAYLIST(p));

    return p->priv->current_position;
}

void playlist_insert_pattern(Playlist* p,
    const gint pos,
    const gint pat)
{
    _playlist_insert_pattern(p, N_("Pattern insertion"), pos, pat, -1);
    editor_select_current_row(p->priv);
}

void
playlist_add_pattern(Playlist* p, const gint copy_from)
{
    PlaylistPriv* priv;

    g_assert(IS_PLAYLIST(p));
    priv = p->priv;
    g_assert(priv->xm != NULL);
    g_assert(!(copy_from > priv->max_pattern));

    if (priv->xm->song_length < priv->max_length) {
        gint n = st_find_first_unused_and_empty_pattern(priv->xm);

        if (n != -1) {
            _playlist_insert_pattern(p,
                copy_from >= 0 ? N_("Pattern add + copy") : N_("Free pattern addition"),
                priv->current_position + 1, n, copy_from);
            playlist_set_position(p, priv->current_position + 1);
        }

        if (copy_from >= 0) {
            st_copy_pattern(&priv->xm->patterns[n], &priv->xm->patterns[copy_from]);
        }
    }
}

/* --- gtk+ object initialization crap --- */

static void
playlist_class_init(PlaylistClass* class)
{
    GObjectClass* g_object_class;

    g_object_class = (GObjectClass*)class;

    playlist_signals[SIG_CURRENT_POSITION_CHANGED] = g_signal_new("current_position_changed",
        G_TYPE_FROM_CLASS(g_object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(PlaylistClass, current_position_changed),
        NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
    playlist_signals[SIG_ENTRY_CHANGED] = g_signal_new("entry_changed",
        G_TYPE_FROM_CLASS(g_object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(PlaylistClass, entry_changed),
        NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
    playlist_signals[SIG_INSERT_PATTERN] = g_signal_new("insert_pattern",
        G_TYPE_FROM_CLASS(g_object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(PlaylistClass, insert_pattern),
        NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
    playlist_signals[SIG_ADD_COPY] = g_signal_new("add_copy",
        G_TYPE_FROM_CLASS(g_object_class),
        (GSignalFlags)G_SIGNAL_RUN_FIRST,
        G_STRUCT_OFFSET(PlaylistClass, add_copy),
        NULL, NULL,
        g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

    class->current_position_changed = NULL;
    class->entry_changed = NULL;
    class->insert_pattern = NULL;
    class->add_copy = NULL;
}

static void
playlist_init(Playlist* p)
{
    p->priv = g_new(PlaylistPriv, 1);

    // These presets should probably be configurable via the interface
    p->priv->max_length = 256;
    p->priv->min_pattern = 0;
    p->priv->max_pattern = 255;

    // A reasonable default playlist
    p->priv->xm = NULL;
    p->priv->current_position = 0;
    p->priv->signals_disabled = 0;
    p->priv->frozen = FALSE;
    p->priv->enabled = TRUE;
    p->priv->skip_logging = FALSE;
    p->priv->row_ins = -1;

    p->priv->dialog = NULL;
}

G_DEFINE_TYPE(Playlist, playlist, GTK_TYPE_EVENT_BOX)
